/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#ifndef __COLLADA_RT__RENDER_H__
#define __COLLADA_RT__RENDER_H__  
		 
#include "Crt/CrtNode.h"
#include "Crt/CrtScene.h"
#include "Crt/CrtCamera.h"
#include "Crt/CrtPhysics.h"
#include "Crt/CrtTexture.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'
   
/**
 * Previously "CrtOrient/CrtNode."
 * This class implements complex instancing of <node>.
 * @see @c RT::Stack.
 */
class Stack_Data
{
COLLADA_(public)
	/**
	 * @c Node is this stack-frame's shared data.
	 *
	 * This could be removed, but it seems as if
	 * it's a useful bit of extra information to
	 * have alongside @c Physics.
	 */
	RT::Node *Node;

	/**
	 * This is used to remove multi-SID Physics
	 * binding candidates. It can be jettisoned
	 * after if passed to @c Update() or if the
	 * root-node is a dummy.
	 */
	RT::Stack_Data *Parent;	

	/**OPAQUE, PRIVATE-DATA
	 * This replaces an old "UpdatingMatrix" flag
	 * while also housing private simulation data.
	 *
	 * @todo
	 * Eventually <instance_physics_model parent>
	 * will have to be represented by this member.
	 */
	btRigidBody *Physics;

COLLADA_(public) 
	/**
	 * Fully animated world matrix for this node.
	 */	
	RT::Matrix Matrix;

  ////THESE SHOULD BE OPTIONAL////

	/**
	 * "InverseTranspose" is a cryptic way to say a normals-rotation matrix.
	 * It's not needed unless <scale> or <skew> or a <matrix> scales/shears.
	 */
	RT::Matrix InverseTransposeMatrix;

COLLADA_(public) //RT::Stack usese these.
	
	Stack_Data():Parent(),Node(),Physics(){}

	int Light(int); void ShowHierarchy_glVertex3d();

	RT::Float *Update_Matrix(RT::Float*,const RT::Matrix&);
};
/**INTERNAL
 * These are sorted according to <geometry> and <controller>.
 * When the <skeleton> of a <skin> is equal it's passed over.
 * EVENTUALLY SORTING SHOULD DO A THIRD CHECK TO ENSURE THIS.
 */
struct Stack_Draw : std::pair<RT::Geometry*,RT::Controller*>
{
	RT::Stack_Data *Data; void *Instance;

	inline RT::Geometry_Instance *AsGeometry()
	{
		return second!=nullptr?nullptr:(RT::Geometry_Instance*)Instance;
	}
	inline RT::Controller_Instance *AsController()
	{
		return second==nullptr?nullptr:(RT::Controller_Instance*)Instance;
	}

	inline std::vector<RT::Material_Instance> &GetMaterials()
	{
		//FYI: There aren't any branches here if Materials is at the same offset.
		return second==nullptr?AsGeometry()->Materials:AsController()->Materials;
	}
};

/**EXPERIMENTAL
 * Trying to get low-level render data out of 
 * @c RT::Frame ("CrtRender.")
 */
class Stack
{
COLLADA_(public) //new APIs
	/**
	 * This replaces the instanced/animated
	 * parts of the old "CrtNode" class. The
	 * elements correspond to nodes. It isn't
	 * necessary to do a hiearchical traversal.
	 */
	std::vector<RT::Stack_Data> Data;
	/**
	 * This holds the instanced/animated node
	 * "transform" values like <rotate> and is
	 * probably going to hold optional matrices
	 * either alongside it or in the back, since
	 * @c RT::Stack_Data doesn't always use every
	 * one of its matrices.
	 */
	std::vector<RT::Float> TransformData;

	/**EXPERIMENTAL
	 * This is sorted according to geometry
	 * instances, also including controllers.
	 */
	std::vector<RT::Stack_Draw> DrawData;
	/**
	 * This is pointers into @c Data for each
	 * <instance_controller> is the @c DrawData.
	 * The pointers are in blocks of bound joints.
	 */
	std::vector<RT::Stack_Data*> ControllerData;

	/**INTERNAL
	 * This rebuilds the "stack." The stack-model
	 * implements the many-level instancing model.
	 */
	void Select(RT::Node *node=nullptr)
	{
		Data.clear(); ControllerData.clear();
		VBuffer1.Clear();
		if(node==nullptr)
		return Data.resize(1); 		
		TransformData.clear();
		Data.reserve(node->CountDescendants());				
		Select_AddData(node,nullptr,0);
		std::sort(DrawData.begin(),DrawData.end());
		Select_AddData_Controllers();		
	}private:
	bool Select_AddData(RT::Node*,RT::Stack_Data*,size_t);
	void Select_AddData_DrawData(void*,void*,RT::Stack_Data*);
	void Select_AddData_Controllers();	

COLLADA_(public) //EXPERIMENTAL
	/**
	 * These buffers may be juggled with @c std::swap. The
	 * buffers are both used if interpolating <controller>
	 * data. The memory may be alternatively used via swap.
	 */
	struct VBuffer
	{			  
		size_t Capacity; 
		/**
		 * @c Memory is for direct memory access.
		 * @c new_Memory is system memory.
		 */
		union{ float *new_Memory,*Memory; };

		void Clear()
		{
			Capacity = 0; 
			COLLADA_RT_array_delete(new_Memory);
			assert(Memory==nullptr);
		}
		
	}VBuffer1,VBuffer2;

COLLADA_(public) //old "CrtRender" APIs

	RT::Material *CurrentMaterial;	
				
	void SetMaterial(RT::Material *mat);
	
	/** 
	 * Set the camera instance we will be rendering from.
	 */
	void ResetCamera();

	//SCHEDULED FOR REMOVAL
	//This is keeping old/lousy code from breaking.
	RT::Stack_Data &SimpleLight();

COLLADA_(public) 
	/**
	 * This version of @c Update() is for starting
	 * physical simulations and animations without
	 * any residual effects. It's likely essential 
	 * for Physics; because the initial states are
	 * just left to chance if not used.
	 */
	void Reset_Update(){ Update(true); }
	/**
	 * This refreshes the world transform matrices.
	 * @see @c Reset_Update().
	 */
	void Update(bool=false);
	
	/**
	 * @c Draw() calls @c Draw_Triangles() internally.
	 * It first sets up lights and may do more passes.
	 */
	void Draw(),Draw_Triangles();
};

/**UNUSED?
 * Getting this out of @c RT::Frame ("CrtRender.")
 */
struct Frame_ShadowMap
{	
	GLuint Id; int Width, Height;

	bool Use, RenderingTo, RenderingWith, Initialized;	

	Frame_ShadowMap():Width(512),Height(512),Id()
	,Use(),RenderingTo(),RenderingWith(),Initialized()
	{}

	void Init();
	void SetupRenderingToShadowMap();
	void SetupRenderingWithShadowMap();	
};

/**
 * Getting this out of @c RT::Frame ("CrtRender.:)
 * @see CrtCommongCg.cpp
 */
struct Frame_Cg
{
	bool Use, Initialized; 
	
	CGcontext Context;

COLLADA_(public)
	/**NEW
	 * This is taken from "CrtGeometry.cpp" and so named
	 * because "fc->resetEffectPassState()" must be done
	 * after drawing/switching materials.
	 */
	void SetPassState(FX::Material*,int);

COLLADA_(public) //old "CrtRender" APIs

	bool Init();
	void Reset();
	bool Destroy();
	bool EnableProfiles(),DisableProfiles(); 
	
	//The rest is mostly clutter.

  #ifdef NDEBUG
  #error Are the "Set" APIs client APIs?
  #endif

  //UNUSED //UNUSED //UNUSED //UNUSED //UNUSED

	bool LoadProgram(int&,const char *fileName, CGprofile);

	int SkinDefaultProgramId;
	int StaticDefaultProgramId;
	int PhongFragmentProgramId;		

	int SkinShadowProgramId;
	int StaticShadowProgramId;
	int PhongFragmentShadowProgramId;
									
	int StaticNormalMapId;
	int SkinNormalMapId; //UNIMPLEMENTED
	int FragmentNormalMapId; 
	
	bool SetDefaultStaticProgram();
	bool SetDefaultSkinProgram();		
	bool SetDefaultFragmentProgram()
	{
		SetPhongFragmentProgram();
	}
	bool SetPhongFragmentProgram();
	
	bool SetShadowMapStaticProgram();
	bool SetShadowMapSkinProgram();
	bool SetShadowMapFragmentProgram();
	
	bool SetNormalMapStaticProgram();
	bool SetNormalMapSkinProgram();
	bool SetNormalMapFragmentProgram();

  //NO IDEA //NO IDEA //NO IDEA //NO IDEA //NO IDEA 
			
	//RT::Stack_Data::Draw() is calling this.
	#ifdef NDEBUG
	#error Is this used on all paths?
	#endif
	float _cgGLSetMatrixParameterfc_matrix[4*4];
	inline void SetCurrentMatrix(RT::Matrix &mat)
	{
		#ifdef NDEBUG
		#error Should this ever transpose?
		#endif
		for(int i=0;i<4*4;i++)
		_cgGLSetMatrixParameterfc_matrix[i] = (float)mat[i];
	}						 
	inline const float *GetCurrentMatrix()
	{
		return _cgGLSetMatrixParameterfc_matrix;
	}
	inline const float *GetCurrentMatrix_InverseTranspose()
	{
		#ifdef NDEBUG
		#error If scaling/shearing this needs to be set.
		#endif
		return _cgGLSetMatrixParameterfc_matrix;
	}
};

/**
 * To limit Get/SetBlah() APIs the idea is to check if
 * these have changed whenever @c Refresh() is invoked.
 */
struct Frame_State : RT::Camera_State
{	
	int Width,Height;	

	//Add these two.
	//bool UseCg;
	//bool UseShadowMap;
	bool UseVBOs;	
	bool UseRender;
	bool UsePhysics;
		
	//Requires Reload
	bool LoadAnimation;
	bool LoadImages;
	bool LoadGeometry;
	bool ShowHierarchy;

	//Animation Controls 
	bool AnimationOn;
	bool AnimationPaused;
	RT::Float Time,Delta;

	/**
	 * Pointer to the active camera instance in the scene.
	 */
	RT::Camera *Camera;
};

/**
 * This tracks the <up_axis> and <unit> states recursively.
 */
struct Frame_Asset
{
	/**
	 * Until there is a class for representing this, these
	 * may be temporarily overriden by the loading process.
	 * @see @c RT::Main_Asset
	 */
	RT::Up Up; RT::Float Meter;
	/**
	 * For animation runtime. (Just extra asset-wide data.)
	 */
	RT::Float FirstKeyTime,LastKeyTime;
					   
	template<class T>
	/** 
	 * Implements @c operator=().
	 */
	void OverrideWith(const daeElement*);
	/** 
	 * Assign @c Up and @c Meter according to 
	 * an <instance_*> reference.
	 */
	void operator=(const daeElement*);	
	/** 
	 * Assign @c Up and @c Meter according to 
	 * an <instance_*> reference.
	 */
	inline void operator=(const DAEP::Element *cp)
	{
		return operator=(dae(cp));
	}
	 
	/** 
	 * @c SetGravity() uses this.
	 */
	void Mult(RT::Float &x, RT::Float &y, RT::Float &z)	
	{	
		x*=Meter; y*=Meter; z*=Meter;
		//   0,M01,  0 
		//-M10,  0,  0 
		//   0,  0,M22
		if(RT::Up::X_UP==Up) std::swap(x*=-1,y);
		//M00,  0,   0
		//  0,  0,-M12
		//  0,M21,   0
		if(RT::Up::Z_UP==Up) std::swap(y*=-1,z);
	}
};

//SCHEDULED FOR REMOVAL
/**
 * Previously "CrtRangeData."
 */
struct Frame_SetRange
{
	RT::Float Min[3],Max[3];
	float Zoom; bool SetFlag;	
	Frame_SetRange(){ ResetZoom(); }	
	void ResetZoom(){ memset(this,0x00,sizeof(*this)); }	
	void SetZoom();

	//Previously "SetRange."
	void operator()(RT::Geometry_Semantic&,size_t);	
};

/**SINGLETON
 * @see @c RT::Main.
 * Previously "CrtRender."
 * @c RT::Main is this type of object. Various parts of the
 * the package assume their @c RT::Frame is @c RT::Main. It
 * is a bad idea to pass @c RT::Main around, since that can
 * create the appearance otherwise, and will make a rewrite
 * less straightforward.
 * @remark "Frame" is like an <iframe> or a picture's frame.
 * This is a humble user interface for the whole RT package.
 */
class Frame : public RT::Frame_State
{
COLLADA_(public)
	/**
	 * This had been "CrtScene *Scene;" 
	 * It should be a daeDB<RT::DBase>.
	 */
	const RT::DBase *const Data; 
	/**
	 * This will be resolved, and it depends on the document.
	 */
	const daeName URL; const daeDOM DOM;

	/**INTERNAL
	 * @c Stack houses the low-level rendering, etc. elements. 
	 */
	RT::Stack Stack; 	

	/**INTERNAL
	 */
	RT::Frame_Asset Asset; RT::Node Scene; RT::Physics Physics;

	#ifdef NDEBUG
	#error Implement RT::Frame_State logic.
	#endif
	/**
	 * This checks the RT::Frame_State members before drawing
	 * each render-frame.
	 */
	bool Refresh();

	/**UNUSED?
	 */
	RT::Frame_Cg Cg; RT::Frame_ShadowMap ShadowMap;	

	/**
	 * Set up @c COLLADA_FX.platform_Filter to target
	 * only specific FX platforms. The default is "PC"
	 * but the FX component prior to 2017 used PC-OGL.
	 * Post-2017 if unset all platform strings are go.
	 */
	FX::Loader COLLADA_FX; RT::Image Missing_Image;
		
	//SCHEDULED FOR REMOVAL
	void Init(),_InitData(),_InitMembers(),_Destroy();	
	/**
	 * Singleton Constructor
	 * It will be a project to support multiple objects
	 * of this type.
	 */
	Frame():Loading(),Data(),URL(""),Missing_Image("default.tga")
	{
		assert(this==&RT::Main); _InitMembers(); 
	}
	~Frame(){ _Destroy(); }
	
	const bool Loading;
	/**
	 * @param index_DAE displaces the contents of @c DOM.
	 * @c Load automatically selects the <scene> element
	 * of a COLLADA document.
	 * @c Loading is set @c true until @c Load completes.
	 * @c URL remembers @c index_DAE's resolved URI data.
	 */
	bool Load(const xs::anyURI &index_DAE);	

	void Unload();

	inline bool Reload(){ return Load(URL); }
		
COLLADA_(public) //old camera management APIs
	/**
	 * These change @c Camera. The cameras loop 
	 * back to the default camera, and so can be
	 * enumerated and managed this way. The state
	 * is shared. The client must keep track of it.
	 */
	void SetDefaultCamera(),SetNextCamera();
		
	#ifdef NDEBUG
	#error Merge this with Camera_State.
	#endif
	RT::Frame_SetRange SetRange;
	inline void ZoomIn(RT::Float zoom)
	{
		ZoomTransform(SetRange.Zoom*0.2*zoom);
	}	
};
//SHORTHAND
static RT::Frame_Asset &Asset = RT::Main.Asset;
//SCHEDULED FOR REMOVAL
/**RAII
 * This class is used to override the <asset> 
 * up/scale. It's not ideal, but it's alright.
 */
struct Main_Asset
{
	const RT::Up Up;
	const RT::Float Meter;	
	Main_Asset(RT::Up up, RT::Float meter)
	:Up(RT::Asset.Up),Meter(RT::Asset.Meter)
	{
		RT::Asset.Up = up; RT::Asset.Meter = meter;
	}
	Main_Asset(const Collada05_XSD::asset *cp)
	:Up(RT::Asset.Up),Meter(RT::Asset.Meter)	
	{
		RT::Asset.Up = cp->up_axis->value->*Up;
		RT::Asset.Meter = cp->unit->meter->*Meter;
	}
	Main_Asset(const DAEP::Element *cp)
	:Up(RT::Asset.Up),Meter(RT::Asset.Meter)
	{
		RT::Asset = cp;
	}
	Main_Asset(const RT::Asset_Index &cp)
	:Up(RT::Asset.Up),Meter(RT::Asset.Meter)
	{
		RT::Asset.Up = RT::GetUp_Meter(cp).first;
		RT::Asset.Meter = RT::GetUp_Meter(cp).second;
	}
	~Main_Asset()
	{
		RT::Asset.Up = Up; RT::Asset.Meter = Meter;
	}
};

//-------.
	}//<-'
}

#endif //__COLLADA_RT__RENDER_H__
/*C1071*/