/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#include <Crt/RT.pch.h> //PCH

#include "Crt/CrtEffect.h"
#include "Crt/CrtLight.h"
#include "Crt/CrtGeometry.h"
#include "Crt/CrtMatrix.h"
#include "Crt/CrtNode.h"
#include "Crt/CrtRender.h"
#include "Crt/CrtTexture.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'
  
void RT::Stack::SetMaterial(RT::Material *mat)
{
	RT::Effect *effect = mat->Effect;
	 //In COLLADA 1.4 a material is required to have a reference to an effect
	assert(effect!=nullptr);

	FLOAT shininess = effect->Shininess;
	//if the scale is 0.0 to 1.0, make it 0.0 to 100.0
	if(shininess<1.0) shininess = shininess*100.0f; 

	//	diffuse.a	=	effect->Transparency;
	glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,&effect->Emission.f0);
	glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,&effect->Ambient.f0);
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,&effect->Diffuse.f0);
	glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,&effect->Specular.f0);
	glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,shininess);

	if(!effect->Textures.empty())
	{
		glEnable(GL_TEXTURE_2D);
		for(size_t i=0;i<effect->Textures.size();i++)
		{
			//glClientActiveTexture(GL_TEXTURE0+i);
			effect->Textures[i]->SetTexture();
		}
	}
	else glDisable(GL_TEXTURE_2D); CurrentMaterial = mat;
};

void RT::Stack::ResetCamera()
{
	RT::Camera *c = RT::Main.Camera;

	//Recalculate the aspect in case the window size has changed.
	#ifdef NDEBUG
	#error COLLADA doesn't have an automatic aspect ratio mode.
	#endif

	//Get the camera from the instance and set the projection matrix from it
	glMatrixMode(GL_PROJECTION); switch(c->Refresh())
	{
	case 2: glOrtho(-c->Xmag,c->Xmag,-c->Ymag,c->Ymag,c->ZNear,c->ZFar); break;
	case 3: gluPerspective(c->Yfov,c->Aspect,c->ZNear,c->ZFar); break;
	}

	//Now get the node from the instance, and set the modelview matrix to the inverse of 
	//the node's matrix.
	glMatrixMode(GL_MODELVIEW);
	RT::Matrix m;
	//NOTE: THIS HAD OVERWRITTEN THE NODE'S MATRIX. FOR NOW CONTROLLING THE CAMERA WON'T.
	RT::MatrixMult(RT::Main.Matrix,RT::Main.Parent->Matrix,m);
	RT::Matrix3x4Invert(m,m);
	#if 2==COLLADA_DOM_PRECISION
	glLoadMatrixd(m);
	#else
	glLoadMatrixf(m);
	#endif
}
		
void RT::Stack::Draw()
{
	//draw lines to each parent node?
	if(RT::Main.ShowHierarchy)
	{	
		//Draw blue lines.
		GLboolean lit = glIsEnabled(GL_LIGHTING);
		if(lit) glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);		
		glColor3f(1,0.5,1); 
		glLineWidth(3); 
		glBegin(GL_LINES);		
		for(size_t i=1;i<Data.size();i++)
		{
			Data[i].ShowHierarchy_glVertex3d();
			Data[i].Parent->ShowHierarchy_glVertex3d();
		}
		glEnd(); glLineWidth(1);
		glEnable(GL_TEXTURE_2D);
		if(lit) glEnable(GL_LIGHTING);
	}	
					
	//reset all lights
	int l=0,lN = 7;
	for(size_t i=0;i<Data.size();i++)
	if(!Data[i].Node->Lights.empty())
	{
		int ll = l; 
		l = std::min(lN,Data[i].Light(l));
		//Enable the light and lighting
		while(ll<l) glEnable(GL_LIGHT0+ll++);
		if(l==lN) break;
	}	
	while(l<lN) glDisable(GL_LIGHT0+l++);
				
	//Are we using shadow maps?
	//!!!GAC Shadow map code may not be functional right now
	if(RT::Main.ShadowMap.Use&&RT::Main.ShadowMap.Initialized)
	{
		//Render the shadow pass
		RT::Main.ShadowMap.SetupRenderingToShadowMap();
		Draw_Triangles();		
		//Now render the scene 
		RT::Main.ShadowMap.SetupRenderingWithShadowMap();
		Draw_Triangles();
	}
	else Draw_Triangles();
}
void RT::Stack::Draw_Triangles()
{	
	RT::Geometry *g = nullptr;

	//HACK: There's not a separate procedure for
	//initializing these vertex-buffers. VBuffer1
	//is cleared when Select() is called.
	if(0==VBuffer1.Capacity)
	{		
		if(DrawData.empty()) return;

		VBuffer2.Clear();
		for(size_t i=0;i<DrawData.size();i++)
		{
			if(g!=DrawData[i].first)
			{	
				g = DrawData[i].first;
				VBuffer1.Capacity = std::max(VBuffer1.Capacity,g->Size_VBuffer());
			}
			if(DrawData[i].second!=nullptr) //3+3 is POSITION+NORMAL.
			{					
				VBuffer2.Capacity = std::max(VBuffer2.Capacity,g->Vertices*(3+3));
			}
		}
		assert(0!=VBuffer1.Capacity);
		VBuffer1.new_Memory = new float[VBuffer1.Capacity];		
		if(0!=VBuffer2.Capacity)
		VBuffer2.new_Memory = new float[VBuffer2.Capacity];
	}
	
	glEnableClientState(GL_VERTEX_ARRAY);

	RT::Stack_Data **cd = nullptr, **cd_end = nullptr;
	if(!ControllerData.empty()) //C++98/03 support
	{
		cd = &ControllerData.front(); cd_end = &ControllerData.back()+1;
	}

	//This is drawing <instance_geometry> and <instance_controller>
	//in shared <geometry> and <skeleton> order. Not material order.
	for(size_t i=0;i<DrawData.size();)	
	{
		RT::Stack_Draw &d = DrawData[i]; g = d.first;
		if(d.second!=nullptr)
		{
			d.AsController()->Advance_ControllerData(cd,cd_end);
			d.first->Fill_VBuffer(VBuffer2.new_Memory);
		}
		else d.first->Fill_VBuffer(VBuffer1.Memory);
		for(;;) //Loop on shared <geometry>.
		{	
			RT::Stack_Draw &d = DrawData[i];
			{
				//HACK: This is a float precision matrix.
				RT::Main.Cg.SetCurrentMatrix(d.Data->Matrix);
				glLoadMatrixf(RT::Main.Cg.GetCurrentMatrix());

				//SHOULD DO THIS HERE. See CrtGeometry.cpp.
				d.first->Draw_VBuffer(VBuffer1.Memory,d.GetMaterials());
			}		
			if(++i<DrawData.size()&&g==DrawData[i].first)
			{
				RT::Stack_Draw &dd = DrawData[i];

				//<geometry> can simply reuse the vbuffer.
				if(dd.second==nullptr||d.second==dd.second
				&&d.AsController()->Joints==dd.AsController()->Joints)
				continue;

				dd.AsController()->Advance_ControllerData(cd,cd_end);
				dd.first->OverrideWith_ControllerData(VBuffer2.new_Memory);				
			}
			else break;
		}
	}
}

void RT::Stack_Data::ShowHierarchy_glVertex3d()
{
	//2017: Any reason this won't work on PlayStation 3?
	//#ifndef SN_TARGET_PS3 
	glVertex3d(Matrix[M30],Matrix[M31],Matrix[M32]);	
	//#endif  
}
int RT::Stack_Data::Light(int l)
{
	const float black[4] = { 0,0,0,0 };

	//This handles the new lighting model, l is the GL light to set up.
	for(size_t i=0;i<Node->Lights.size();i++,l++)
	{
		RT::Light *light = Node->Lights[i];		
		
		//There's only one color in the light, try to set the GL parameters sensibly.		
		if(RT::Light_Type::AMBIENT==light->Type)		
		{
			glLightfv(GL_LIGHT0+l,GL_AMBIENT,&light->Color.r);
			glLightfv(GL_LIGHT0+l,GL_DIFFUSE,black);
			glLightfv(GL_LIGHT0+l,GL_SPECULAR,black);	
		}
		else
		{
			glLightfv(GL_LIGHT0+l,GL_AMBIENT,black);
			glLightfv(GL_LIGHT0+l,GL_DIFFUSE,&light->Color.r);
			glLightfv(GL_LIGHT0+l,GL_SPECULAR,&light->Color.r);	
		}

		//Get the light position from the light instance's parent node's Matrix.
		FX::Float4 position(Matrix[M30],Matrix[M31],Matrix[M32],1);
		if(RT::Light_Type::DIRECTIONAL!=light->Type)
		{
			glLightfv(GL_LIGHT0+l,GL_POSITION,&position.x);		
			glLightf(GL_LIGHT0+l,GL_CONSTANT_ATTENUATION,light->ConstantAttenuation);
			glLightf(GL_LIGHT0+l,GL_LINEAR_ATTENUATION,light->LinearAttenuation);
			glLightf(GL_LIGHT0+l,GL_QUADRATIC_ATTENUATION,light->QuadraticAttenuation);
		}

		if(RT::Light_Type::POINT!=light->Type)
		{
			RT::MatrixTransform(InverseTransposeMatrix,FX::Float3(0,0,1),position);

			if(RT::Light_Type::DIRECTIONAL==light->Type)
			{
				position.w = 0; glLightfv(GL_LIGHT0+l,GL_POSITION,&position.x);
			}
			else
			{
				position.w = 1; glLightfv(GL_LIGHT0+l,GL_SPOT_DIRECTION,&position.x);
			}
		}

		if(RT::Light_Type::SPOT==light->Type)
		{
			//The Conformance Test Suite suggests /2 (or a full angle.)
			glLightf(GL_LIGHT0+l,GL_SPOT_CUTOFF,light->FalloffAngle/2);
			glLightf(GL_LIGHT0+l,GL_SPOT_EXPONENT,light->FalloffExponent);			
		}
		else glLightf(GL_LIGHT0+l,GL_SPOT_CUTOFF,180); //Necessary?
	}

	return l;
}

/*REFERENCE
int RT::Frame::GenerateVBO()
{
	if(UseVBOs&&VBOsAvailable)
	{
		size_t vboID;
		//Generate And Bind The Vertex Buffer
		glGenBuffers(1,&vboID);

		return vboID;
	}
	return -1;
}
bool RT::Frame::_UpdateVBO(GLenum type, GLuint vboID, const void *data, GLsizeiptrARB size)
{
	if(UseVBOs&&VBOsAvailable&&vboID>0)
	{
		//Bind The Buffer
		glBindBuffer(type,vboID);
		//Load The Data
		glBufferData(type,size,data,GL_STATIC_DRAW);

		return true;
	}
	return false;
}
bool RT::Frame::BindVBO(GLenum type, GLuint vboID)
{
	if(UseVBOs&&VBOsAvailable)
	{
		glBindBuffer(type,vboID);
		return true;
	}
	else return false;
}
void RT::Frame::FreeVBO(size_t vboUID)
{
	//Delete VBO
	if(UseVBOs&&vboUID>0)
	{
		//Free The VBO Memory
		glDeleteBuffers(1,&vboUID);	assert(vboUID!=0);
	}
} */

void RT::Frame_ShadowMap::Init()
{
	if(Initialized) return;

	//setup the depth texture
	glGenTextures(1,&Id);
	glBindTexture(GL_TEXTURE_2D,Id);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_COMPARE_MODE_ARB,GL_COMPARE_R_TO_TEXTURE_ARB);
	glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT,Width,Height,0,GL_DEPTH_COMPONENT,GL_UNSIGNED_INT,nullptr);

	Initialized = true;
}

void RT::Frame_ShadowMap::SetupRenderingToShadowMap()
{
	//set tc flag so polygroups will know which shader to use 
	RenderingTo = true;
	RenderingWith = false;

	//gl stuff to setup rendering to the shadow map 
	glEnable(GL_CULL_FACE);
	glViewport(0,0,Width,Height);
	glClear(GL_DEPTH_BUFFER_BIT);
	glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.1f,4);

	//void RT::Frame::PushLightViewMatrix()
	{
		//should get the current light from the scene and get it 
		//position and what now but for just for now going to 
		//set a hard value 
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		//Not implemented yet: need to look into perspective range
		gluPerspective(35,(double)Width/Height,1,30);

		//setup light view
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		//light is directly above for now 
		gluLookAt(0,30,0, //eye (light)  position
		0,0,0, //target
		0,1,0); //up vector
	}
}	 
void RT::Frame_ShadowMap::SetupRenderingWithShadowMap()
{
	//daeEH::Verbose<<"Setting Rendering with shadowmap.";

	//set tc flag so polygroups will know which shader to use 
	RenderingWith = true;
	RenderingTo = false;

	glDisable(GL_CULL_FACE);
	glDisable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0,0);
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
	glBindTexture(GL_TEXTURE_2D,Id);
	
	glViewport(0,0,Width,Height);
	glClear(GL_DEPTH_BUFFER_BIT);

	//void RT::Frame::PopLightViewMatrix()
	{
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

	//need to get the current camera and reset it again because we have since over written 
	//the model view setup by SceneUpdate
	RT::Main.Stack.ResetCamera();
}

//-------.
	}//<-'
}

/*C1071*/