/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2003-2006 Erwin Coumans  http://continuousphysics.com/Bullet/

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/

#include <Crt/RT.pch.h> //PCH
  
#include "Crt/CrtPhysics.h"
#include "Crt/CrtRender.h"
#include "Crt/CrtGeometry.h"

#if 2==COLLADA_DOM_PRECISION
#define BT_USE_DOUBLE_PRECISION
#endif

#ifdef NO_BULLET
class btTransform;
class btTypedConstraint;
class btVector3;
#else //http://bullet.sf.net Erwin Coumans
#include "btBulletDynamicsCommon.h"
#include "LinearMath/btQuickprof.h"
#include "LinearMath/btIDebugDraw.h"
#include "LinearMath/btTransform.h"
#include "LinearMath/btVector3.h"
#include "LinearMath/btDefaultMotionState.h"
#include "BulletCollision/CollisionShapes/btShapeHull.h"
#include "BulletCollision/CollisionShapes/btBoxShape.h"
#include "BulletCollision/CollisionShapes/btSphereShape.h"
#include "BulletCollision/CollisionShapes/btCylinderShape.h"
#include "BulletCollision/CollisionShapes/btConeShape.h"
#include "BulletCollision/CollisionShapes/btStaticPlaneShape.h"
#include "BulletCollision/CollisionShapes/btConvexHullShape.h"
#include "BulletCollision/CollisionShapes/btTriangleMesh.h"
#include "BulletCollision/CollisionShapes/btConvexTriangleMeshShape.h"
#include "BulletCollision/CollisionShapes/btTriangleMeshShape.h"
#include "BulletCollision/CollisionShapes/btCompoundShape.h"
#ifdef SPU_BULLET //PlayStation 3
#include "SpuDispatch/btPhysicsEffectsWorld.h"//currently in Bullet/SpuDispatch
#include "BulletCollision/BroadphaseCollision/btDbvtBroadphase.h"
#include "SpuDispatch/BulletDebugDrawer.h"
#include "BulletMultiThreaded/SequentialThreadSupport.h"
#include "BulletMultiThreaded/SequentialThreadSupport.h"
#include "BulletMultiThreaded/SpuGatheringCollisionDispatcher.h"
#include "SpuDispatch/BulletCollisionSpursSupport.h"
#include "SpuDispatch/BulletConstraintSolverSpursSupport.h"
#include "SpuDispatch/btPhysicsEffectsWorld.h"
#include "Physics/TaskUtil/SpursTask.h"
#include "Physics/TaskUtil/spurs_util_spu_printf_service.h"
#define SPU_THREAD_GROUP_PRIORITY 250
#define SPURS_THREAD_PRIORITY 1000
#define SPURS_NAME "PhysicsEffects"
#define NUM_MAX_SPU 3
#endif //SPU_BULLET //PlayStation 3
#endif //NO_BULLET 

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'

void RT::Physics::Update(double delta_time)
{
	if(bt.dynamicsWorld==nullptr) return;
	//bt.dynamicsWorld->updateAabbs();
		
	#ifdef NDEBUG
	#error Is btMotionState reset when the simulation is reset?
	#endif

	#ifndef NO_BULLET		
	//float fixedTimeStep = 1.f/60.f;
	//if you want to test some physics simulation, uncomment out next line
	//delta_time = fixedTimeStep;
	bt.dynamicsWorld->stepSimulation(delta_time);
	#endif
}	 

void RT::Physics::SetGravity(RT::Float x, RT::Float y, RT::Float z)
{
	#ifndef NO_BULLET		
	bt.dynamicsWorld->setGravity(btVector3(x,y,z));
	#endif
	daeEH::Verbose<<"Set gravity to "<<x<<","<<y<<","<<z;
}

#ifdef NO_BULLET //EXPERIMENTAL
#define using_namespace_NO_BULLET using namespace CrtPhysics;
#else
#define using_namespace_NO_BULLET
#endif
namespace CrtPhysics
{
	//EXPERIMENTAL
	typedef struct NO_BULLET_shape
	{
		NO_BULLET_shape(...){}
		void addChildShape(){}
	}btCompoundShape,btCollisionShape
	,btStaticPlaneShape,btBoxShape,btSphereShape,btCylinderShapeZ
	,btShapeHull,btConvexHullShape;

	struct transform : RT::Matrix
	{
		operator btTransform&()
		{
			return *(btTransform*)this;
		}
		transform(const transform&);
		transform(const daeContents &content)
		{
			//This is untested with X/Z_UP.
			//Scaling probably will not work here.
			MatrixLoadAsset(*this,RT::Asset.Up);
			content.for_each_child(*this);
			//Assuming BT layout!
			if(RT::Matrix_is_COLLADA_order) RT::MatrixTranspose(*this);
		}
		void operator()(const xs::const_any &e)
		{
			RT::Matrix &m = *this; RT::Float x,y,z,a;
			
			switch(e->getElementType()) //ORDER-IS-IMPORTANT
			{
			case DAEP::Schematic<Collada05::rotate>::genus:		
			{
				Collada05::const_rotate t = e->a<Collada05::rotate>();
				if(t!=nullptr&&4==t->value->get4at(0,x,y,z,a))
				MatrixRotateAngleAxis(m,x,y,z,a);
				break;		
			}
			case DAEP::Schematic<Collada05::translate>::genus:		
			{
				Collada05::const_translate t = e->a<Collada05::translate>();
				if(t!=nullptr&&3==t->value->get3at(0,x,y,z))
				RT::MatrixTranslate(m,x,y,z);
				break;	
			}}
		}
	};
}

void RT::Physics::Bind_instance_rigid_constraint
(int constrain1, int constrain2, Collada05::const_rigid_constraint rigid_constraint)
{	
	btRigidBody *body1 = 0, *body2 = 0;
	#ifdef NDEBUG
	#error This is an exponential lookup.
	#endif
	int many1 = 0, many2 = 0;
	for(size_t i=0;i<RT::Main.Stack.Data.size();i++)
	{	
		btRigidBody *body = RT::Main.Stack.Data[i].Physics;
		if(body==nullptr) continue;

		if((void*)constrain1==body->getUserPointer())
		{
			many1++; body1 = body;
		}
		if((void*)constrain2==body->getUserPointer())
		{
			many2++; body2 = body;
		}
	}
	if(1!=many1||1!=many2)
	{
		daeEH::Error<<"Too few or too many of one or both rigid bodies "
		<<rigid_constraint->ref_attachment->rigid_body<<"("<<many1<<"),"
		<<rigid_constraint->attachment->rigid_body<<"("<<many2<<") for constraint "<<rigid_constraint->sid<<"\n"
		<<"(Expected both to be 1. Are many-to-many semantics called for?)";
		return;
	}

	Collada05::const_limits limits = rigid_constraint->technique_common->limits;
	btVector3 linearMin(0,0,0),linearMax(0,0,0),angularMin(0,0,0),angularMax(0,0,0);
	if(limits!=nullptr)
	{
		Collada05::const_min min,max;
		min = limits->linear->min; max = limits->linear->max;
		if(min!=nullptr) min->value->get3at(0,linearMin[0],linearMin[1],linearMin[2]);
		if(max!=nullptr) max->value->get3at(0,linearMax[0],linearMax[1],linearMax[2]);

		min = limits->swing_cone_and_twist->min; max = limits->swing_cone_and_twist->max;		
		if(min!=nullptr) min->value->get3at(0,angularMin[0],angularMin[1],angularMin[2]);
		if(max!=nullptr) max->value->get3at(0,angularMax[0],angularMax[1],angularMax[2]);
	}
	
	//Assuming the same <asset> space applies to both attachments.
	//It can be interpreted as using each attachment's body/node??
	CrtPhysics::transform 
	ref_transform(rigid_constraint->ref_attachment->content),
	transform(rigid_constraint->attachment->content);

	const bool useReferenceFrameA = true;
	btGeneric6DofConstraint *c = bt.New<btGeneric6DofConstraint>
	(*body1,*body2,ref_transform,transform,useReferenceFrameA);
			
	//convert INF / -INF into lower > upper			
	const RT::Float linearCheckThreshold = 999999;
	const RT::Float angularCheckThreshold = 180; //check this											   
	//free means upper < lower, 
	//locked means upper == lower
	//limited means upper > lower
	//limitIndex: first 3 are linear, next 3 are angular		
	for(size_t i=0;i<3;i++)
	{
		if(linearMin[i]<-linearCheckThreshold
		 ||linearMax[i]>+linearCheckThreshold)
		{
			//disable limits
			linearMin[i] = 1; linearMax[i] = 0;
		}

		if(angularMin[i]<-angularCheckThreshold
		 ||angularMax[i]>+angularCheckThreshold)
		{
			//disable limits
			angularMin[i] = 1; angularMax[i] = 0;
		}
	}
	c->setLinearLowerLimit(linearMin);
	c->setLinearUpperLimit(linearMax);
	c->setAngularLowerLimit(angularMin);
	c->setAngularUpperLimit(angularMax); bt.dynamicsWorld->addConstraint(c); 	
}

void RT::Physics::Bind_instance_rigid_body(int constraint, RT::RigidBody &out, RT::Stack_Data &in)
{	
	RT::Float mass = out.Mass;
	btVector3 localInertia(0,0,0);
	if(!out.Dynamic&&mass!=0)
	{
		//This is an implementation detail.
		//daeEH::Warning<<"Non-dynamic objects need to have zero mass!";
		mass = 0;
	}
	if(out.Dynamic&&mass==0)
	{
		//This is an implementation detail.
		//daeEH::Warning<<"Dynamic rigidbodies need nonzero mass!";
		mass = 1;
	}	
	if(mass!=0)
	out.Shape->calculateLocalInertia(mass,localInertia);
	
	struct _m : btMotionState
	{
		RT::Matrix &m; _m(RT::Matrix &m):m(m){}

		static void copy(RT::Float *dst, const RT::Float *src)
		{
			if(Matrix_is_COLLADA_order)
			MatrixTranspose(*(RT::Matrix*)dst,*(RT::Matrix*)src);
			else MatrixCopy(*(RT::Matrix*)dst,*(RT::Matrix*)src);
		}
		virtual void getWorldTransform(btTransform &wt)const
		{
			daeCTC<sizeof(wt)==sizeof(RT::Matrix)>();
			copy(&wt.getBasis()[0][0],m);
		}
		virtual void setWorldTransform(const btTransform &wt)
		{
			copy(m,&wt.getBasis()[0][0]);
		}
	}*ms = bt.New<_m>(in.Matrix);
	//using motionstate is recommended, it provides interpolation capabilities,
	//and only synchronizes 'active' objects
	btRigidBody *body = bt.New<btRigidBody>(mass,ms,out.Shape,localInertia);
	
	//This is an opaque identifier for attaching constraints.
	//It's hierarchical.
	body->setUserPointer((void*)constraint);

	bt.dynamicsWorld->addRigidBody(body); in.Physics = body;
}

void RT::RigidBody::_LoadShape(Collada05::const_rigid_body::technique_common in)
{
	using_namespace_NO_BULLET //EXPERIMENTAL

	struct RT::Physics::bt &bt = RT::Main.Physics.bt;

	//A setLocalScaling API implements <unit meter> last.
	//1 makes CrtPhysics::transform not scale <translate>.
	const RT::Float m = RT::Asset.Meter;
	if(m!=1) RT::Asset.Meter = 1; //POINT-OF-NO-RETURN
	bool requires_compound = RT::Asset.Up!=RT::Up::Y_UP;	

	btCompoundShape *compound = nullptr;
	btCollisionShape *current = nullptr;
	Collada05::const_rigid_body::technique_common::shape shape;
	for(size_t i=0;i<in->shape.size();i++)
	{
		shape = in->shape[i]; current = nullptr;

		if(!shape->plane.empty())
		{
			Collada05::const_equation plane = shape->plane->equation;
			if(plane!=nullptr)
			{
				RT::Float a,b,c,d;
				if(4==plane->value->get4at(0,a,b,c,d))
				current = bt.New<btStaticPlaneShape>(btVector3(a,b,c),d);
			}
		}
		else if(!shape->box.empty())
		{
			Collada05::const_half_extents box = shape->box->half_extents;			
			RT::Float x,y,z;
			if(box!=nullptr&&3==box->value->get3at(0,x,y,z))
			current = bt.New<btBoxShape>(btVector3(x,y,z));
		}
		else if(!shape->sphere.empty())
		{
			RT::Float radius = shape->sphere->radius->value->*RT::Float(-1);
			if(radius>0)			
			current = bt.New<btSphereShape>(radius);
		}
		else if(!shape->cylinder.empty())
		{
			Collada05::const_cylinder cylinder = shape->cylinder;
			RT::Float radius1 = 0, radius2, height = 0;
			if(!cylinder->radius.empty())
			{
				switch(cylinder->radius->value->get2at(0,radius1,radius2))
				{
				case 1: radius2 = radius1; case 2: break; default: continue;
				}
				height = cylinder->height->value->*RT::Float(-1);
			}
			if(height>=0&&radius1>0&&radius2>0)			
			current = bt.New<btCylinderShapeZ>(btVector3(radius1,height,radius2));
		}
		else if(!shape->instance_geometry.empty())
		{
			xs::anyURI URI = shape->instance_geometry->url;
			linked_geometry:
			Collada05::const_geometry geom = URI.get<Collada05::geometry>();			
			URI = geom->convex_mesh->convex_hull_of->*"";
			if(!URI.empty())
			{
				URI.setParentObject(geom);
				daeEH::Verbose<<"Linked geometry: "<<URI.getURI();
				//In theory convex_hull_of can go on indefinitely.
				goto linked_geometry; 
			}
			RT::Geometry *g = 
			const_cast<RT::DBase*>(RT::Main.Data)->LoadGeometry(geom);
			if(g==nullptr) continue;

			#ifdef NDEBUG
			#error Use a table to not duplicate mesh-based shapes.
			#endif

			if(Dynamic&&geom->convex_mesh.empty())
			{
				#ifdef NDEBUG
				#error Take a look at ConvexDecompositionDemo.cpp?
				#endif
				daeEH::Verbose<<"Assuming Physics <mesh> is sufficiently convex.\n"
				"(Decomposition support may be added later on if it is practical.)";
			}
			
			const size_t s = g->Positions.Stride;
			const size_t o = g->Positions.Offset;
			const RT::Float *v0 = &g->Positions->value[o];
			//UNUSED: There might be reasons to turn this back on later.
			if(0/*in.triangles.size*/) 
			{
				//2017: This path was coded for <mesh> with <triangles>.
				//It's been rewritten for preservation. But documentation
				//seems to advise against this approach.
				//Note: This formulation includes all of the mesh's faces.

				btTriangleMesh *trimesh = bt.New<btTriangleMesh>();
				
				for(size_t i=0;i<g->Elements.size();i++)
				if(GL_TRIANGLES==g->Elements[i].Mode)
				{
					GLuint *e = &g->ElementBuffer[g->Elements[i].Region];
					GLuint *et = e+g->Elements[i].Width;
					for(;e<et;e+=3)
					{
						const RT::Float *v;
						v = v0+s*e[0];
						btVector3 x(v[0],v[1],v[2]);
						v = v0+s*e[1];
						btVector3 y(v[0],v[1],v[2]);
						v = v0+s*e[2];
						btVector3 z(v[0],v[1],v[2]);
						trimesh->addTriangle(x,y,z);
					}
				}				

				if(Dynamic)
				{
					daeEH::Warning<<"Moving concave mesh not supported, transformed into convex.";
					current = bt.New<btConvexTriangleMeshShape>(trimesh);
				}
				else
				{
					daeEH::Verbose<<"Static concave triangle mesh added.";
					bool useQuantizedAabbCompression = false;
					current = bt.New<btBvhTriangleMeshShape>(trimesh,useQuantizedAabbCompression);
				}
			}
			else //BUILD A POINT-CLOUD HULL (HOW ACCURATE CAN IT BE?)
			{
				btConvexHullShape *convexHullShape = bt.New<btConvexHullShape>();

				for(size_t i=0,iN=i+s*g->Vertices;i<iN;i+=s)	
				convexHullShape->addPoint(btVector3(v0[i],v0[i+1],v0[i+2]));	
				daeEH::Verbose<<"Created convexHullShape with "<<g->Vertices<<" points.";
				current = convexHullShape;
			}
			
			//2017: The BT wiki says advises to not use large meshes.
			//http://www.bulletphysics.org/mediawiki-1.5.8/index.php/Collision_Shapes
			if(Dynamic&&g->Vertices>100)
			{
				assert(current->isConvex());
				btConvexShape *currently_convex = (btConvexShape*)current;

				//http://www.bulletphysics.org/mediawiki-1.5.8/index.php/BtShapeHull_vertex_reduction_utility
				btShapeHull hull(currently_convex);
				btScalar margin = current->getMargin(); hull.buildHull(margin);

				//MEMORY-LEAK? The code in the link about doesn't delete the input shape.
				//http://bulletphysics.org/Bullet/BulletFull/btShapeHull_8cpp_source.html
				bt.Delete_back(current); //delete current;
				const btScalar *bt_New_compat = (btScalar*)hull.getVertexPointer();
				int bt_New_compat2 = hull.numVertices();
				current = bt.New<btConvexHullShape>(bt_New_compat,bt_New_compat2);

				assert(hull.numVertices()<=100&&hull.numVertices()<(int)g->Vertices);
				daeEH::Warning<<"Reduced large convexHullShape to "<<hull.numVertices()<<" points.";
			}
		}

		//compound if more then 1 shape, or a non-identity local shapetransform
		if(current!=nullptr)
		if(requires_compound||!shape->rotate.empty()||!shape->translate.empty())
		{
			if(compound==nullptr)
			{
				requires_compound = true; compound = bt.New<btCompoundShape>();
			}
			compound->addChildShape(CrtPhysics::transform(shape->content),current);			
		}	 
		else requires_compound = true;
	}

	Shape = compound!=nullptr?compound:current;
	
	if(m!=1) //Restore to original status.
	{
		RT::Asset.Meter = m; 
		if(Shape!=nullptr) Shape->setLocalScaling(btVector3(m,m,m));
	}
}
	
RT::Physics::Physics()
{
	const float world = 1000; assert(RT::Physics::OK);
	#ifndef NO_BULLET	
	///collision configuration contains default setup for memory, collision setup
	bt.collisionConfiguration = bt.New<btDefaultCollisionConfiguration>();
	#ifdef SPU_BULLET //PlayStation 3
	//----------------------------------------------------------
	//SPURS instance
	CellSpursAttribute attributeSpurs;
	int ret = cellSpursAttributeInitialize(&attributeSpurs,NUM_MAX_SPU,SPU_THREAD_GROUP_PRIORITY,SPURS_THREAD_PRIORITY,false);
	if(ret!=0) daeEH::Verbose<<"cellSpursAttributeInitialize failed: "<<ret;
	if(ret!=0) return;
	ret = cellSpursAttributeSetNamePrefix(&attributeSpurs,SPURS_NAME,strlen(SPURS_NAME));
	if(ret!=0) daeEH::Verbose<<"cellSpursAttributeSetNamePrefix failed: "<<ret;
	if(ret!=0) return;
	ret = cellSpursInitializeWithAttribute(&bt.SpursInstance,&attributeSpurs);
	if(ret!=0) daeEH::Verbose<<"cellSpursInitializeWithAttribute failed "<<ret;
	if(ret!=0) return;
	//----------------------------------------------------------
	//SPURS printfserver
	ret = sampleSpursUtilSpuPrintfServiceInitialize(&bt.SpursPrintfService,&bt.SpursInstance,SPURS_THREAD_PRIORITY);
	if(ret!=0) daeEH::Verbose<<"spurs_printf_service_initialize failed: "<<ret;
	if(ret!=0) return;
	int numSpuTasks = NUM_MAX_SPU;
	bt.collisionThreadSupport = bt.New<BulletCollisionSpursSupport>(&bt.SpursInstance,numSpuTasks,numSpuTasks);
	bt.dispatcher = bt.New<SpuGatheringCollisionDispatcher>(bt.collisionThreadSupport,numSpuTasks,bt.collisionConfiguration);
	#else
	bt.dispatcher = bt.New<btCollisionDispatcher>(bt.collisionConfiguration);
	#endif	
	bt.pairCache = bt.New<btAxisSweep3>
	(btVector3(-world,-world,-world),btVector3(world,world,world));
	bt.constraintSolver = bt.New<btSequentialImpulseConstraintSolver>();
	bt.dynamicsWorld = bt.New<btDiscreteDynamicsWorld>
	(bt.dispatcher,bt.pairCache,bt.constraintSolver,bt.collisionConfiguration);
	return;
	daeCTC<Physics::OK
	&&sizeof(btScalar)==sizeof(RT::Float)
	&&sizeof(btTransform)==sizeof(RT::Matrix)>();
	#endif //NO_BULLET
}

daeOK RT::Physics::Snapshot(const xs::anyURI &remote_directory)
{	
	xs::anyURI base = RT::Main.URL;
	base.erase(base.getURI_filenameCP());		
	base.setIsResolved(); remote_directory.resolve();
	if(base==remote_directory)
	{
		//Writing to the same directory is not done.
		assert(0); return false; 
	}

	//This is really just protecting the ref-counts.
	daeArray<xs::any> undo;

	//Ready a snapshot, saving the old "transforms."	
	for(size_t i=0;i<RT::Main.Stack.Data.size();i++)
	{	
		RT::Stack_Data &data = RT::Main.Stack.Data[i];
		btRigidBody *body = data.Physics;
		if(body==nullptr) continue;

		//SCHEDULED FOR REMOVAL
		Collada05_XSD::node *node = 
		(Collada05_XSD::node*)data.Node->Fragment;
		daeMeta &meta = dae(node)->getMeta();
		for(size_t i=0;i<data.Node->Transforms.size();i++)
		{
			//SCHEDULED FOR REMOVAL
			undo.push_back((DAEP::Element*)data.Node->Transforms[i].Fragment);
			meta.removeWRT(dae(node),undo.back());
		}

	#ifdef NDEBUG
	#error //// Do this with ColladaLAB.inl /////////////
	#endif

		RT::Quaternion q;
		RT::Matrix &m = data.Matrix;

		//First, translation.
		q[0] = m[M30]; q[1] = m[M31]; q[2] = m[M32];
		if(data.Parent!=nullptr)
		{
			//Convert to local translation.
			RT::Matrix &mm = data.Parent->Matrix;
			q[0]-=mm[M30]; q[1]-=mm[M31]; q[2]-=mm[M32];
		}
		//if(dot(q)>3*SIMD_EPSILON*SIMD_EPSILON)
		node->translate->value->assign(q,3);

		//Solve for local rotation.
		RT::MatrixToQuat(m,q);
		if(data.Parent!=nullptr)
		{
			RT::Quaternion qq;
			RT::MatrixToQuat(data.Parent->Matrix,qq);
			for(int i=0;i<3;i++) 
			qq[i] = -qq[1]; //invert (unit conjugate)
			#ifdef NDEBUG
			#error A*B or B*A?
			#endif
			QuaternionMult(qq,q); 
		}
		
		//Convert to angle representation.
		q[3] = RT::Float(2*acos(q[3])*SIMD_DEGS_PER_RAD);
		if(q[3]>SIMD_EPSILON)
		{
			RT::Float dot = q[0]*q[0]+q[1]*q[1]+q[2]*q[2];
			if(dot>3*SIMD_EPSILON*SIMD_EPSILON)
			{
				//Convert to axis representation.
				for(int i=0;i<3;i++) 
				q[i]*=1/sqrt(dot);			
				node->rotate->value->assign(q,4);
			}
		}
	}

	//Note, this a debug feature for developers. It
	//isn't this package's place to write documents.
	daeOK OK = DAE_ERROR; xs::anyURI URI;
	daeDOM &DOM = const_cast<daeDOM&>(RT::Main.DOM);
	for(size_t i=0,iN=DOM.getDocs().size();i<iN;i++)	
	{
		const xs::anyURI &rel = DOM.getDoc(i)->getDocURI();
		if(!rel.transitsURI(base)
		 ||!URI.setURI(remote_directory,rel.data()+base.size()))
		continue;
		URI.setIsResolved();

		Collada05::COLLADA COLLADA = DOM.getDoc<Collada05::COLLADA>(i);
		if(COLLADA==nullptr) 
		continue;
		
		Collada05::contributor contributor = ++COLLADA->asset->contributor;

		contributor->author->value = 
		"http://bullet.sf.net Erwin Coumans";
		contributor->authoring_tool->value =
		#ifdef WIN32
		"Bullet ColladaPhysicsViewer-Win32-0.8";
		#elif defined(__APPLE__)
		"Bullet ColladaPhysicsViewer-MacOSX-0.8";
		#else
		"Bullet ColladaPhysicsViewer-UnknownPlatform-0.8";
		#endif
		contributor->comments->value = 
		"Comments to Physics Forum at http://www.continuousphysics.com/Bullet/phpBB2/index.php";

		OK = DOM.getDoc(i)->writeTo(URI); if(!OK) break;
	}

	//Undo the "transform" changes made by the snapshot.
	for(size_t t=0,i=0;i<RT::Main.Stack.Data.size();i++)
	{	
		RT::Stack_Data &data = RT::Main.Stack.Data[i];
		if(data.Physics==nullptr) continue;

		//SCHEDULED FOR REMOVAL
		Collada05_XSD::node *node = 
		(Collada05_XSD::node*)data.Node->Fragment;
		node->translate = ""; 
		node->rotate = "";
		daeMeta &meta = dae(node)->getMeta();
		for(size_t i=0;i<data.Node->Transforms.size();i++)
		meta.placeWRT(dae(node),undo[t++]);
	}	

	return OK;
}

//-------.
	}//<-'
}

/*C1071*/