/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#include <Crt/RT.pch.h> //PCH

#include "Crt/CrtCg.h"
#include "Crt/CrtRender.h"
#include "Crt/CrtLight.h"
#include "Crt/CrtGeometry.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'

extern std::vector<char> CrtTexture_buffer;

//!!!GAC temporary WINDOWS ONLY for performance timing
#ifdef _WIN32 
LARGE_INTEGER update_time,render_time;
#endif

//#ifdef SN_TARGET_PS3
//#include <sys/raw_spu.h>
//#include <sys/sys_time.h>
//#elif !defined(_WIN32)
//#include <sys/time.h>
//#endif
static RT::Float CrtRender_timer()
{
	#ifdef SN_TARGET_PS3
	{
		system_time_t t = sys_time_get_system_time();
		return t*RT::Float(0.000001);
	}
	#elif defined(_WIN32)
	{
		return timeGetTime()*RT::Float(0.001);
	}
	#else
	{
		struct timeval LTV;
		struct timezone LTZ;

		if(gettimeofday(&LTV,&LTZ)!= -1)
		{
			long long LTime = LTV.tv_sec;

			LTime*=1000000;
			LTime+=LTV.tv_usec;
			unsigned long time = LTime&0xFFFFFFFF;
			return (RT::Float)time/1000000;
		}
	}
	#endif
	return 0;
}

static void CrtRender_UpdateDelta()
{	
	//this function should be called once per frame 	
	static RT::Float fps = 0;
	static RT::Float time = 0;
	static RT::Float oldTime = CrtRender_timer();
	static int nbrFrames = 0;
	static bool UpdatedOnce = false;
	const RT::Float updateTime = 1;

	time = CrtRender_timer();

	nbrFrames++;

	if(time>(oldTime+updateTime))
	{
		fps = (RT::Float)nbrFrames;
		RT::Main.Delta = 1.0f/(fps)*updateTime;
		//If Delta is running less than 15 fps, stop animation and wait till fps come back to above 15 fps
		//if(RT::Main.Delta>0.04f) 
		//RT::Main.Delta = 0; 

		oldTime = time;
		nbrFrames = 0;

		UpdatedOnce = true;
		#if 0 
		daeEH::Verbose<<"FPS is "<<fps<<", Time is "<<time;
		#endif
		#ifdef _WIN32  //!!!GAC temporary windows only performance timing code
		LARGE_INTEGER frequency;
		QueryPerformanceFrequency(&frequency);
		//daeEH::Verbose<<"update "<<(double)update_time.QuadPart/(double)frequency.QuadPart<<
		//", render "<<(double)render_time.QuadPart/(double)frequency.QuadPart<<
		//", frequency "<<frequency.QuadPart;
		#endif
	}

	//until first update delta = 0;
	if(!UpdatedOnce) RT::Main.Delta = 0;
}

//This is a table of <up_axis> & <unit> value pairs.
static std::vector<RT::Up_Meter> 
CrtRender_Up_Meter(1,std::make_pair(RT::Up::Y_UP,1));
const RT::Up_Meter &RT::GetUp_Meter(size_t i)
{
	return CrtRender_Up_Meter[i];
}
RT::Asset_Index::Asset_Index()
{
	RT::Up_Meter cmp(RT::Asset.Up,RT::Asset.Meter);
	for(size_t i=0;i<CrtRender_Up_Meter.size();i++)	
	if(cmp==CrtRender_Up_Meter[i]) 
	{
		_i = i&0xFF; return;
	}	
	_i = 0xFF&CrtRender_Up_Meter.size();
	CrtRender_Up_Meter.push_back(cmp); 
}
void RT::Frame_Asset::operator=(const daeElement *e)
{
	if(e==nullptr) 
	{
		Up = RT::Up::Y_UP; Meter = 1; return;
	}
	operator=(e->getParent()); switch(e->getElementType())
	{
	#define _(x) \
	case DAEP::Schematic<Collada05::x>::genus:\
		OverrideWith<Collada05::x>(e); break;
	_(COLLADA)
	_(library_nodes)_(node)	
	_(library_visual_scenes)_(visual_scene)
	_(library_physics_scenes)_(physics_scene)
	_(library_physics_models)_(physics_model)
	//Assuming <source><asset> doesn't specify up/meter.
	_(library_geometries)_(geometry)
	#undef _
	default: assert(0);
	}
}
template<class T>
void RT::Frame_Asset::OverrideWith(const daeElement *e)
{
	T &t = (T&)e; if(!t->asset.empty())
	{
		Up = t->asset->up_axis->value->*Up;
		Meter = t->asset->up_axis->value->*Meter;
	}
} 

void RT::Frame::Unload()
{
	Missing_Image.DeleteTexture();
	Physics.Clear(); Scene.Clear(); 	
	Stack.Select();
	delete Data; _InitData(); 
	const_cast<daeDOM&>(DOM).clear_of_content();
	Cg.Reset();
	UsePhysics = true;
}
void RT::Frame::_Destroy()
{
	Unload(); if(Cg.Initialized) Cg.Destroy();
}

bool RT::Frame::Load(const xs::anyURI &URI)
{
	Unload(); assert(nullptr==Data);

	//Create a new scene and name it the same as the file being read.
	(void*&)Data = COLLADA_RT_new(RT::DBase);
	
	//in case of multithreaded loading 
	const_cast<bool&>(Loading) = true;

	//daeEH::Verbose<<"Loading file "<<fileName;

	//Setup the type of renderer and initialize Cg if necessary (we may need the context for loading)
	//!!!GAC this code used to come after the load, but now the load needs a Cg context.
	if(UseRender)
	{
		//try and initialize cg if we can as set the default shaders 
		if(Cg.Use) Cg.Init();

		if(ShadowMap.Use) ShadowMap.Init();
	}
	else Cg.Use = UseVBOs = ShadowMap.Use = false;

	daeEH::Verbose<<"COLLADA_DOM Load Started.";	

	daeDocRoot<Collada05::COLLADA> res = 
	const_cast<daeDOM&>(DOM).openDoc<Collada05::COLLADA>(URI);
	if(res!=DAE_OK)
	{
		daeEH::Error<<
		"Failed to load document:\n"<<URI.getURI();
		Unload();
		daeEH::Error<<"Error loading the COLLADA file:"
		"\n"<<URI.getURI()<<"\n"
		"Make sure it is COLLADA 1.4.1 or greater.";			
		const_cast<bool&>(Loading) = false;
		 return false;
	}
	else const_cast<daeName&>(URL) = res->getDocURI().getURI(); 

	daeEH::Verbose<<"COLLADA_DOM Runtime database initialized from:"
	"\n"<<URL;
	
	//SCHEDULED FOR REMOVAL
	const_cast<RT::DBase*>(Data)->LoadCOLLADA(res);
	
	CrtTexture_buffer.swap(std::vector<char>());
	
	//in case of multithreaded loading 
	const_cast<bool&>(Loading) = false;

	daeEH::Verbose<<"Done Loading "<<URL;	

	Time = RT::Asset.FirstKeyTime;

	RT::Main.SetRange.ResetZoom(); return true;
}

bool RT::Frame::Refresh()
{
	if(Data==nullptr) return false;
		   
	//daeEH::Verbose<<"Rendering Data."; 

	//Update the scene, this traverses all the nodes and makes sure the transforms are up-to-date
	//It also sets up the hardware lights from the light instances in the scene.  (note the hardware
	//lights aren't used if you're rendering using Cg shaders.
	#ifdef _WIN32  //!!!GAC temporary windows only performance timing code
	LARGE_INTEGER temp_time;
	QueryPerformanceCounter(&temp_time);
	#endif
	{	
		//too keep animation playback stabl
		//no matter the frame rate 
		CrtRender_UpdateDelta();
	
		if(!AnimationPaused)
		{
			if(UsePhysics)
			Physics.Update(Delta);		
		}
		Stack.Update();

		#ifdef NDEBUG
		#error This is old code, sorely in need of review.
		#endif
		//update animation if not paused 
		if(!AnimationPaused)
		{
			Time+=Delta;
			if(Time>RT::Asset.LastKeyTime)
			Time = RT::Asset.FirstKeyTime+(Time-RT::Asset.LastKeyTime);
		}
		//set not to use the animation if 
		//animation is turned off 
		if(!AnimationOn)
		Time = -1;
	}
	#ifdef _WIN32  //!!!GAC temporary windows only performance timing code
	QueryPerformanceCounter(&update_time);
	update_time.QuadPart = update_time.QuadPart-temp_time.QuadPart;
	QueryPerformanceCounter(&temp_time);
	#endif

	if(UseRender)
	{
		SetRange.SetZoom();
		//Set the camera instance we will render from.
		Stack.ResetCamera();
		Stack.Draw();
	}

	//!!!GAC temporary windows only performance timing code
	#ifdef _WIN32  
	QueryPerformanceCounter(&render_time);
	render_time.QuadPart = render_time.QuadPart-temp_time.QuadPart;
	#endif

	return true;
}

//!!!GAC This is a temporary fix to be removed by the refactoring, the contents of InitMembers and Init
//!!!GAC used to be all in one Init function that was called by the constructor AND by the mainlines
//!!!GAC of most of the samples.  This was causing Init to be called twice.  This wasn't a problem till
//!!!GAC we wanted to add Cg initialization which shouldn't be called in the global constructor and
//!!!GAC shouldn't be called twice.  To fix the problem I split the init function so the constructor
//!!!GAC now only initializes members and Init initializes members and CG, this avoids the need to
//!!!GAC change every sample that calls Init.  I will clean this up later.
void RT::Frame::_InitData()
{
	const_cast<daeName&>(URL) = "";	
	const_cast<RT::DBase*&>(Data) = nullptr; 
}
void RT::Frame::_InitMembers()
{
	_InitData(); //2017

	Cg.Use = false;
	UseVBOs = false;
	UseRender = true;
	UsePhysics = true;
	  
	Cg.Initialized = false;
	Cg.Context = nullptr;

	Cg.SkinDefaultProgramId = -1;
	Cg.StaticDefaultProgramId = -1;
	Cg.PhongFragmentProgramId = -1;

	Cg.SkinShadowProgramId = -1;
	Cg.StaticShadowProgramId = -1;
	Cg.PhongFragmentShadowProgramId = -1;

	Cg.StaticNormalMapId = -1;
	Cg.SkinNormalMapId = -1;
	Cg.FragmentNormalMapId = -1;

	Width = 640;
	Height = 480;
	Delta = 0.03f;

	Asset.Up = RT::Up::Y_UP; Asset.Meter = 1;

	LoadAnimation = true;
	LoadImages = true;
	LoadGeometry = true;
	ShowHierarchy = false;

//2017: Moved from old CrtScene
	
	//for animation runtime only
	Time = Asset.LastKeyTime = Asset.FirstKeyTime = 0;

	//Animation Controls 
	AnimationOn = true;	AnimationPaused = false;

	//Memory guarantees
	Stack.Select(); SetDefaultCamera();

	Stack.CurrentMaterial = nullptr;	
}

void RT::Frame::Init()
{
	_InitMembers();

	//!!!GAC The new crt render path requires Cg to be initialized before anything is loaded
	Cg.Init();

	daeEH::Verbose<<
	"------------------------------\n"
	"--- COLLADA-RT Initialized ---\n"
	"------------------------------";
}
		  
void RT::Frame::SetDefaultCamera()
{
	assert(!Stack.Data.empty());

	Parent = &Stack.Data[0]; 
	static RT::Camera default_camera;
	default_camera.Id = "COLLADA_RT_default_camera";
	Camera = &default_camera;
}
void RT::Frame::SetNextCamera()
{
	size_t i = Parent-&Stack.Data[0];
	while(i<Stack.Data.size())
	{
		std::vector<RT::Camera*> &ic = Stack.Data[i].Node->Cameras;
		if(ic.empty()) continue;

		size_t j = 0;
		if(Parent==&Stack.Data[i])
		{
			for(j=0;j<ic.size();j++)		
			if(Camera==ic[j])
			break;
			if(j<=ic.size()-1) continue;
		}
		Parent = &Stack.Data[i]; Camera = ic[j];
	}
	if(i>=Stack.Data.size()) SetDefaultCamera(); 

	daeEH::Verbose<<"Active camera instance on node "<<
	Parent->Node->Id<<" is based on camera "<<Camera->Id;	
}

//SCHEDULED FOR REMOVAL
//This is keeping old/lousy code from breaking.
RT::Stack_Data &RT::Stack::SimpleLight()
{
	//Give preference to scenes without light.
	if(!Data[0].Node->Lights.empty())
	return Data[0];

	size_t ambient = 0; //ambient
	for(size_t i=0;i<Data.size();i++)
	{
		std::vector<RT::Light*> &il = Data[i].Node->Lights;
		if(il.empty()) 
		continue;
		if(il[0]->Type!=RT::Light_Type::AMBIENT)
		return Data[i];
		ambient = i; 
		if(il.size()==1)
		continue;
		//Assuming client is not tracking this.
		std::swap(il[0],il[1]); 
		return Data[i];
	}
	//Assuming ambient is as good as any positioned at 0,0,0.
	if(0==ambient)
	{
		static RT::Light default_light;
		default_light.Id = "COLLADA_RT_default_light";
		default_light.Type = RT::Light_Type::AMBIENT;
		Data[0].Node->Lights.push_back(&default_light);
	}
	return Data[ambient];
}

void RT::Frame_SetRange::operator()(RT::Geometry_Semantic &positions, size_t vertices)
{
	//This seems to be set once on the first time the new document is drawn.
	if(SetFlag) return;

	size_t s = positions.Stride;
	const RT::Float *it = &positions->value[positions.Offset];
	for(const RT::Float *itt=it+s*vertices;it<itt;it+=s)
	{
		for(size_t i=0;i<3;i++) if(it[i]>Max[i]) Max[i] = it[i];
		for(size_t i=0;i<3;i++) if(it[i]<Min[i]) Min[i] = it[i];
	}
	float max_hight = (float)std::max(Max[1],-Min[1]);
	float zoom_height = max_hight*1.376f; //1.376 = 1/tan(36);
	float this_zoom = 3*std::max((float)Max[2],zoom_height);
	Zoom = std::max(Zoom,this_zoom);	
	//daeEH::Verbose<<"max_height is "<<max_hight<<
	//", zoom_height is "<<zoom_height<<
	//", this_zoom is "<<this_zoom<<
	//", zoom is "<<Zoom;
};
void RT::Frame_SetRange::SetZoom()
{
	if(SetFlag) return;
	//Disabling this so there's no chance it prints "Zoom is 0" every frame.
	/*if(Zoom!=0)*/ SetFlag = true;

	RT::Main.ZoomTransform(Zoom);

	daeEH::Verbose<<"Zoom is "<<Zoom;
}
  
bool RT::Stack::Select_AddData(RT::Node *p, RT::Stack_Data *pp, size_t depth)
{
	//The second test is concerned with controllers.
	if(depth>COLLADA_RT_MAX_DEPTH)
	{
		daeEH::Error<<"COLLADA_RT_MAX_DEPTH reached. Truncating graph.";
		return false;
	}
	if(Data.size()==Data.capacity()) 
	{
		daeEH::Error<<"Truncated graph capacity reached. Truncating controller.";
		return false;
	}

	//Reminder: Don't reject nodes with SIDs or IDs.
	//The Physics bindings might need to trace them.

	Data.push_back(RT::Stack_Data());

	RT::Stack_Data &d = Data.back(); d.Node = p; d.Parent = pp;
		
	for(size_t i=0;i<p->Geometries.size();i++)	
	Select_AddData_DrawData(&p->Geometries[i],nullptr,&d);	
	for(size_t i=0;i<p->Controllers.size();i++)
	Select_AddData_DrawData(nullptr,&p->Controllers[i],&d);

	for(size_t i=0;i<p->Nodes.size();i++) 
	Select_AddData(p->Nodes[i],&d,depth+1); return true;
}
void RT::Stack::Select_AddData_DrawData(void *g, void *c, RT::Stack_Data *d)
{
	RT::Stack_Draw draw; 
	draw.Data = d; draw.Instance = g==nullptr?g:c;
	if(g!=nullptr) draw.first = draw.AsGeometry()->Geometry;
	if(c!=nullptr) draw.second = draw.AsController()->Controller;
	if(c!=nullptr) draw.first = draw.second->Geometry;
	DrawData.push_back(draw);
}
void RT::Stack::Select_AddData_Controllers()
{
	//<instance_controller> cannot create an instance of nodes.
	//Therefore if a controller must be independently animated
	//or bound to a physical system, it must be uniquely bound.
	for(size_t i=0;i<DrawData.size();)
	{
		RT::Stack_Draw &d = DrawData[i++];
		RT::Controller_Instance *ic = d.AsController();
		while(i<DrawData.size()&&d.second==DrawData[i].second)
		{
			//Skip instances of controllers sharing a skeleton.
			if(ic==nullptr||ic->Joints==DrawData[i].AsController()->Joints)
			i++;
			else break;
		}
		if(ic==nullptr) continue;

		for(size_t j,i=ic->Skeletons;i<ic->Joints.size();i++)
		{
			RT::Node *p = ic->Joints[i];
			for(j=0;j<Data.size();j++) if(p==Data[j].Node)
			{
				ControllerData.push_back(&Data[j]);
				break;
			}
			if(j==Data.size())
			{
				if(!Select_AddData(p,&Data[0],0))
				return; //Error?
				ControllerData.push_back(&Data.back());					
			}
		}		
	}
}

void RT::Stack::Update(bool resetting)
{
	//This resets physics simulations and 
	//intializes the transform data, so if
	//a node is transformed, it just has to
	//be reset.
	if(resetting) 
	{
		//This can go in the main loop, but
		//it makes it hard to see what's what.
		TransformData.clear();		
		for(size_t i=1;i<Data.size();i++)
		{
			std::vector<RT::Transform> &v = Data[i].Node->Transforms;						
			for(size_t j,i=0;i<v.size();i++)
			{
				for(j=0;j<v[i].Size;j++)
				TransformData.push_back(v[i].ResetData[j]);	
				if(j==16&&!RT::Matrix_is_COLLADA_order)
				RT::MatrixTranspose(*(RT::Matrix*)(&TransformData.back()-15));
			}
		}
		if(TransformData.empty()) //C++98/03
		TransformData.resize(1);
	}
	RT::Float *td = &TransformData[0]; //C++98/03

	size_t up_i = -1; 
	RT::Matrix &asset = Data[0].Matrix;		
	for(size_t i=1;i<Data.size();i++)
	{
		RT::Stack_Data &d = Data[i];
		if(up_i!=d.Node->Asset)
		{
			up_i = d.Node->Asset;
			RT::Up_Meter up_m = RT::GetUp_Meter(up_i);
			RT::MatrixLoadAsset(asset,up_m.first,up_m.second);
		}
		if(d.Physics!=nullptr||resetting)
		td = d.Update_Matrix(td,asset);
		else for(size_t i=0;i<d.Node->Transforms.size();i++)
		td+=d.Node->Transforms[i].Size;
	}
	if(up_i!=0) RT::MatrixLoadIdentity(asset);

	#ifdef NDEBUG
	#error This is not usually required/should be intelligent.
	#endif
	for(size_t i=1;i<Data.size();i++)
	{
		RT::Stack_Data &d = Data[i];
		d.InverseTransposeMatrix[M30] = 0;
		d.InverseTransposeMatrix[M31] = 0;
		d.InverseTransposeMatrix[M32] = 0;
		d.InverseTransposeMatrix[M33] = 1;
		RT::Matrix3x4Invert(d.Matrix,d.InverseTransposeMatrix);
		RT::MatrixTranspose(d.InverseTransposeMatrix);
	}
}
RT::Float *RT::Stack_Data::Update_Matrix(RT::Float *td, const RT::Matrix &up_axis_and_unit_meter)
{
	RT::Transform *tf;
	for(size_t i=0;i<Node->Animators.size();i++)
	Node->Animators[i].Animate(td);

	RT::Matrix LocalMatrix(up_axis_and_unit_meter);		
	for(size_t i=0;i<Node->Transforms.size();i++,td+=tf->Size)	
	switch((tf=&Node->Transforms[i])->Type)
	{
	case RT::Transform_Type::Rotate:

		RT::MatrixRotateAngleAxis(LocalMatrix,td[0],td[1],td[2],td[3]);
		break;

	case RT::Transform_Type::Translate: 

		one_third_of_lookat: //REMOVE ME
		RT::MatrixTranslate(LocalMatrix,td[0],td[1],td[2]);
		break;

	case RT::Transform_Type::Scale:

		RT::MatrixScale(LocalMatrix,td[0],td[1],td[2]);
		break;

	case RT::Transform_Type::Matrix:

		RT::MatrixMult(*(RT::Matrix*)td,LocalMatrix);
		break;

	default:
	//case RT::Transform_Type::Skew:
	//case RT::Transform_Type::LookAt:
		#ifdef NDEBUG
		#error Fix this.
		#endif
		assert(0); goto one_third_of_lookat;
		//daeEH::Error<<"Programming error. Not expecting this kind of Transform."; //Do Later
		break;
	}

	assert(Parent!=nullptr);
	RT::MatrixMult(LocalMatrix,Parent->Matrix,Matrix); return td;
} 

//-------.
	}//<-'
}

/*C1071*/